﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Security.Principal;
using System.Text;

namespace SVMctl_I2C_CSharp
{

    // SVM, SVOボード制御用クラス
    public static class SVIControl
    {
        /// <summary>
        /// Device Driverオープンフラグ
        /// </summary>
        private static bool driverOpenFlag = false;

        /// <summary>
        /// SVM,SVOシリーズの初期化処理
        /// </summary>
        /// <param name="deviceId">複数のボードが接続された場合、0から昇順で各基板に付与される番号</param>
        /// <param name="boardType">初期化対象のボード種別指定。SVM, SVIシリーズの場合0、SVOシリーズの場合1</param>
        public static void DeviceOpen(int deviceId, int boardType)
        {
            // 初期化対象基板が既にOpenされている場合クローズし、フラグ修正
            DeviceClose();
            
            // 制御ライブラリ初期化
            NativeMethods.Init();

            // 接続先を指定しDevice Driverをオープン
            UInt32 ret = NativeMethods.OpenEx(NativeMethods.APP_WHO_CTL, deviceId, boardType);

            // Device Driverオープンフラグ設定
            if (ret == NativeMethods.RET_NORMAL)
            {
                driverOpenFlag = true;
            }
        }


        /// <summary>
        /// 接続済デバイスのクローズ処理
        /// </summary>
        public static void DeviceClose()
        {
            if (driverOpenFlag == true)
            {
                NativeMethods.Close(NativeMethods.APP_WHO_CTL);
                driverOpenFlag = false;
            }

            // 制御ライブラリ終了処理
            NativeMethods.End();
        }


        /// <summary>
        /// 指定したdeviceId, boardTypeに対してDevice Driverのオープンを試み、OK(0x1) or ERROR(null)を返す
        /// </summary>
        /// <param name="deviceId">複数のボードが接続された場合、0から昇順で各基板に付与される番号</param>
        /// <param name="boardType">初期化対象のボード種別指定。SVM, SVIシリーズの場合0、SVOシリーズの場合1</param>
        /// <returns>正常終了:NativeMethods.RET_NORMAL(0x1), ERROR:null</returns>
        public static UInt32? DeviceConnectionCheck(int deviceId, int boardType)
        {
            UInt32? ret;

            // 初期化対象基板が既にOpenされている場合クローズし、フラグ修正
            DeviceClose();

            NativeMethods.Init();
            ret = NativeMethods.OpenEx(NativeMethods.APP_WHO_CTL, deviceId, boardType);
            driverOpenFlag = true;

            if (ret != NativeMethods.RET_NORMAL)
            {
                DeviceClose();
                return null;
            }

            DeviceClose();
            return ret;
        }


        /// <summary>
        /// 制御ライブラリのバージョン情報をstring型で取得
        /// </summary>
        /// <returns>正常終了:NativeMethods.RET_NORMAL(0x1), ERROR:null</returns></returns>
        public static string GetDllVersion()
        {
            StringBuilder dllVersion = new StringBuilder(256);
            UInt32? ret = NativeMethods.GetVersion(dllVersion);

            if (ret == NativeMethods.RET_NORMAL)
            {
                return dllVersion.ToString();
            }
            else
            {
                return null;
            }
        }


        /// <summary>
        /// 指定したボード上のFPGA(Artix7/Spartan6)の内部レジスタの値を読み出す
        /// </summary>
        /// <param name="deviceId">複数のボードが接続された場合、0から昇順で各基板に付与される番号</param>
        /// <param name="boardType">初期化対象のボード種別指定。SVM, SVIシリーズの場合0、SVOシリーズの場合1</param>
        /// <param name="address">レジスタアドレス</param>
        /// <returns>正常終了:UInt32型レジスタ値, ERROR:null</returns></returns>
        public static UInt32? FpgaRegisterRead(int deviceId, int boardType, UInt32 address, bool checkIsUsb3 = false, bool boardInitEnd = true)
        {
            // FPGAレジスタ読み込みAPI使用時の固定値(unCommandId, ucSSId)
            byte unCommdId = 0x7;
            if (checkIsUsb3 == true) { unCommdId = 0x9; }

            const byte ucSSId = 0x99;

            // FPGAレジスタ読み込みは4byte単位固定
            const int len = 0x4;
            byte[] receiveBuffer = new byte[len];

            // boardInitEndがtrueの場合、ボード指定で初期化処理
            if (boardInitEnd == true)
            {
                DeviceOpen(deviceId, boardType);
            }

            // 受信データ格納用ポインタ
            IntPtr receivePointer = Marshal.AllocCoTaskMem(len);

            // FPGAレジスタ読み込みAPI実行
            UInt32? ret = NativeMethods.FpgaRegisterRead(unCommdId, ucSSId, address, (UInt32)len, receivePointer);

            // アンマネージドメモリポインター(DLL側)からマネージド配列(C#側)にデータをコピー
            Marshal.Copy(receivePointer, receiveBuffer, 0, len);

            // 受信データ格納用メモリ空間を開放
            Marshal.FreeCoTaskMem(receivePointer);

            // boardInitEndがtrueの場合、接続済みデバイスのクローズ処理
            if (boardInitEnd == true)
            {
                DeviceClose();
            }

            // 正常に通信が行えた場合、4byteの受信データを32bit幅UIntに格納し、戻り値として返す
            if (ret == NativeMethods.RET_NORMAL)
            {
                ret = (UInt32)(receiveBuffer[0] << 8 * 3 | receiveBuffer[1] << 8 * 2 | receiveBuffer[2] << 8 * 1 | receiveBuffer[3]);
                return ret;
            }

            // APIからの戻り値(通信エラー有無)がNativeMethods.RET_NORMALでない時、nullを返す
            else
            {
                return null;
            }
        }


        /// <summary>
        /// 指定したボード上のFPGA(Artix7/Spartan6)の内部レジスタに対し値を書き込み
        /// </summary>
        /// <param name="deviceId">複数のボードが接続された場合、0から昇順で各基板に付与される番号</param>
        /// <param name="boardType">初期化対象のボード種別指定。SVM, SVIシリーズの場合0、SVOシリーズの場合1</param>
        /// <param name="addrress">レジスタアドレス</param>
        /// <param name="data">書き込みデータ</param>
        /// <returns>正常終了:UInt32型レジスタ値, ERROR:null</returns></returns>
        public static UInt32? FpgaRegisterWrite(int deviceId, int boardType, UInt32 addrress, UInt32 data, bool checkIsUsb3 = false, bool boardInitEnd = true)
        {
            // FPGAレジスタ書き込みAPI使用時の固定値(unCommandId, ucSSId)
            byte unCommdId = 0x8;
            if (checkIsUsb3 == true) { unCommdId = 0x0A; }

            const byte ucSSId = 0x99;

            // FPGAレジスタ読み込みは4byte単位固定
            const int len = 0x4;
            byte[] sendBuffer = new byte[len];

            // 送信データ格納用ポインタ
            IntPtr sendPointer = Marshal.AllocCoTaskMem(len);

            // boardInitEndがtrueの場合、ボード指定で初期化処理
            if (boardInitEnd == true)
            {
                DeviceOpen(deviceId, boardType);
            }

            // 引数のデータ4バイトのバイト配列に格納
            for (int i = 0; i < len; i++)
            {
                sendBuffer[i] = (byte)((data >> 8 * (3 - i)) & 0xFF);
            }

            // マネージド配列(C#側)からアンマネージドメモリポインター(DLL側)にデータをコピー
            Marshal.Copy(sendBuffer, 0, sendPointer, len);

            // FPGAレジスタ書き込みAPI実行
            UInt32 ret = NativeMethods.FpgaRegisterWrite(unCommdId, ucSSId, addrress, (UInt32)len, sendPointer);

            // 送信データ格納用メモリ空間を開放
            Marshal.FreeCoTaskMem(sendPointer);

            // boardInitEndがtrueの場合、接続済みデバイスのクローズ処理
            if (boardInitEnd == true)
            {
                DeviceClose();
            }

            // APIからの戻り値(通信エラー有無)がNativeMethods.RET_NORMALでない時、nullを返す
            if (ret == NativeMethods.RET_NORMAL)
            {
                return ret;
            }
            else
            {
                return null;
            }
        }


        /// <summary>
        /// 指定したボードの基板上FW, FPGAバージョン読み込み
        /// </summary>
        /// <param name="deviceId">複数のボードが接続された場合、0から昇順で各基板に付与される番号</param>
        /// <param name="boardType">初期化対象のボード種別指定。SVM, SVIシリーズの場合0、SVOシリーズの場合1</param>
        /// <returns>正常終了:配列長4のbyte配列</returns>
        public static byte[] SVMVersionInfo(int deviceId, int boardType, bool boardInitEnd = true)
        {
            // 基板上FW, FPGAバージョン読み込みは4byte単位固定
            const int len = 4;
            byte[] receiveBuffer = new byte[len];

            // boardInitEndがtrueの場合、ボード指定で初期化処理
            if (boardInitEnd == true)
            {
                DeviceOpen(deviceId, boardType);
            }

            // 受信データ格納用ポインタ
            IntPtr ptr = Marshal.AllocCoTaskMem(len);

            // 基板上FW, FPGAバージョン読み込みAPI実行
            UInt32 ret = NativeMethods.SVMVersionInfo((UInt32)len, ptr);

            // アンマネージドメモリポインター(DLL側)からマネージド配列(C#側)にデータをコピー
            Marshal.Copy(ptr, receiveBuffer, 0, len);

            // 受信データ格納用メモリ空間を開放
            Marshal.FreeCoTaskMem(ptr);

            if (ret == NativeMethods.RET_NORMAL)
            {
                // boardInitEndがtrueの場合、接続済みデバイスのクローズ処理
                if (boardInitEnd == true)
                {
                    DeviceClose();
                }
                return receiveBuffer;
            }
            else
            {
                // boardInitEndがtrueの場合、接続済みデバイスのクローズ処理
                if (boardInitEnd == true)
                {
                    DeviceClose();
                }
                return null;
            }
        }
        
        public static FWInfo GetRevision(int deviceId, int boardType, bool boardInitEnd = true)
        {
            // boardInitEndがtrueの場合、ボード指定で初期化処理
            if (boardInitEnd == true)
            {
                DeviceOpen(deviceId, boardType);
            }

            
            int[] revision = new int[1]; //IntPtrの相互変換の都合上int[]に
            byte[] versionInfo = new byte[3];
            byte[] customCode = new byte[4];
            
            IntPtr ptr_rev_num = Marshal.AllocCoTaskMem(Marshal.SizeOf(typeof(int)));
            IntPtr ptr_versionInfo = Marshal.AllocCoTaskMem(3);
            IntPtr ptr_customCode = Marshal.AllocCoTaskMem(4);
            
            var ret = NativeMethods.SVI05API_GetRevision(ptr_rev_num, ptr_versionInfo, ptr_customCode);
            
            Marshal.Copy(ptr_rev_num, revision, 0, 1);
            Marshal.Copy(ptr_versionInfo, versionInfo, 0, 3);
            Marshal.Copy(ptr_customCode, customCode, 0, 4);
            
            // 送信データ格納用メモリ空間を開放
            Marshal.FreeCoTaskMem(ptr_versionInfo);
            Marshal.FreeCoTaskMem(ptr_customCode);
            Marshal.FreeCoTaskMem(ptr_rev_num);
            
            if (ret == NativeMethods.RET_NORMAL)
            {
                // boardInitEndがtrueの場合、接続済みデバイスのクローズ処理
                if (boardInitEnd == true)
                {
                    DeviceClose();
                }
                String tmpCustomCode = ((char)customCode[0]).ToString() + ((char)customCode[1]).ToString() + ((char)customCode[2]).ToString() + ((char)customCode[3]).ToString();
                return new FWInfo(revision[0], versionInfo[0], versionInfo[1], versionInfo[2], tmpCustomCode);
            }
            else
            {
                // boardInitEndがtrueの場合、接続済みデバイスのクローズ処理
                if (boardInitEnd == true)
                {
                    DeviceClose();
                }
                return new FWInfo();
            }
            
        }


        /// <summary>
        /// I2C経由で指定したデバイスからデータを読み出し
        /// </summary>
        /// <param name="deviceId">複数のボードが接続された場合、0から昇順で各基板に付与される番号</param>
        /// <param name="boardType">初期化対象のボード種別指定。SVM, SVIシリーズの場合0、SVOシリーズの場合1</param>
        /// <param name="slaveAddress">対象デバイスのI2C Slave Address</param>
        /// <param name="subAddress">対象デバイス内の読み込み先Sub Address</param>
        /// <param name="dataLen">読み込みデータのバイト長</param>
        /// <param name="wordAccess">ワードアクセス(16bit Sub Address)有無</param>
        /// <returns></returns>
        public static byte[] I2C_BlockRead(int deviceId, int boardType, UInt32 slaveAddress, UInt32 subAddress, UInt32 dataLen, int sAddrType, bool boardInitEnd = true)
        {
            // 読み込みバイト長(4byte未満の場合もAPIには4byteと指定)
            UInt32 len = 4;
            // 読み込みバイト長が4byteより大きい場合、実際の値をlenに代入
            if (dataLen > 4) { len = dataLen; }

            // 受信バッファの最小byte長は4byte
            byte[] receiveBuffer = new byte[len];

            // 戻り値格納用バッファの最小byte長は1byte (引数で指定されたdataLenと一致)
            byte[] ret_value = new byte[dataLen];

            // boardInitEndがtrueの場合、ボード指定で初期化処理
            if (boardInitEnd == true)
            {
                DeviceOpen(deviceId, boardType);
            }

            // 受信データ格納用ポインタ
            IntPtr ptr = Marshal.AllocCoTaskMem((int)len);

            // I2C Block Read API実行
            UInt32 ret = NativeMethods.I2CBlockRead(slaveAddress, subAddress, len, ptr, sAddrType);

            // アンマネージドメモリポインター(DLL側)からマネージド配列(C#側)にデータをコピー
            Marshal.Copy(ptr, receiveBuffer, 0, (int)len);

            // 受信データ格納用メモリ空間を開放
            Marshal.FreeCoTaskMem(ptr);

            // 正常に通信が行えた場合、読み取りデータをバイト配列で返す
            if (ret == NativeMethods.RET_NORMAL)
            {
                // 受信バッファから戻り値格納用バッファへdataLenのbyte長のデータをコピー
                Array.Copy(receiveBuffer, 0, ret_value, 0, dataLen);

                // boardInitEndがtrueの場合、接続済みデバイスのクローズ処理
                if (boardInitEnd == true)
                {
                    DeviceClose();
                }

                return ret_value;
            }
            else
            {
                // boardInitEndがtrueの場合、接続済みデバイスのクローズ処理
                if (boardInitEnd == true)
                {
                    DeviceClose();
                }
                return null;
            }
        }


        /// <summary>
        /// I2C経由で指定したデバイスへデータを書き込み
        /// </summary>
        /// <param name="deviceId">複数のボードが接続された場合、0から昇順で各基板に付与される番号</param>
        /// <param name="boardType">初期化対象のボード種別指定。SVM, SVIシリーズの場合0、SVOシリーズの場合1</param>
        /// <param name="slaveAddress">対象デバイスのI2C Slave Address</param>
        /// <param name="subAddress">対象デバイス内の書き込み先Sub Address</param>
        /// <param name="data">書き込みデータ</param>
        /// <param name="wordAccess">ワードアクセス(16bit Sub Address)有無</param>
        /// <returns></returns>
        public static UInt32? I2C_BlockWrite(int deviceId, int boardType, UInt32 slaveAddress, UInt32 subAddress, byte[] data, int sAddrType, bool boardInitEnd = true)
        {
            // 引数の書き込みデータのbyte長をlenに代入
            int len;
            if (data == null)
            {
                return null;
            }
            else
            {
                len = data.Length;
            }

            // 送信データ格納用ポインタ
            IntPtr ptr = Marshal.AllocCoTaskMem(len);

            // boardInitEndがtrueの場合、ボード指定で初期化処理
            if (boardInitEnd == true)
            {
                DeviceOpen(deviceId, boardType);
            }

            // マネージド配列(C#側)からアンマネージドメモリポインター(DLL側)にデータをコピー
            Marshal.Copy(data, 0, ptr, len);

            // I2C Block Write API実行
            UInt32 ret = NativeMethods.I2CBlockWrite(slaveAddress, subAddress, (UInt32)len, ptr, sAddrType);

            // 送信データ格納用メモリ空間を開放
            Marshal.FreeCoTaskMem(ptr);

            // 正常に通信が行えた場合、NativeMethods.RET_NORMAL(0x1)を返す
            if (ret == NativeMethods.RET_NORMAL)
            {
                // boardInitEndがtrueの場合、接続済みデバイスのクローズ処理
                if (boardInitEnd == true)
                {
                    DeviceClose();
                }

                return ret;
            }
            else
            {
                // boardInitEndがtrueの場合、接続済みデバイスのクローズ処理
                if (boardInitEnd == true)
                {
                    DeviceClose();
                }

                return null;
            }
        }


        /// <summary>
        /// 現在接続されている基板のコレクションを最大16件まで取得します。
        /// </summary>
        /// <returns>現在接続されている基板のコレクション</returns>
        public static List<BoardInformation> GetBoardInformationList()
        {
            // 取得するボードの最大数
            const int boardMax = 16;
            // 戻り値
            List<BoardInformation> boardInformationList = new List<BoardInformation>();
            // BoardType
            int boardType;

            // 基板情報取得およびコレクション追加処理
            for (int i = 0; i < boardMax; i++)
            {
                // 基板情報
                BoardInformation boardInformation = new BoardInformation();

                // SVMまたはSVIシリーズとしてアクセスできた場合
                if (DeviceConnectionCheck(i, 0) == NativeMethods.RET_NORMAL)
                {
                    boardType = 0;
                }
                // SVOシリーズとしてアクセスできた場合
                else if (DeviceConnectionCheck(i, 1) == NativeMethods.RET_NORMAL)
                {
                    boardType = 1;
                }
                else
                {
                    continue;
                }
                // V1.1.0.0 Vendorモードのボードが接続された場合の例外処理追加
                try
                {
                    boardInformation.FwInfo = SVIControl.GetRevision(i, boardType);
                }
                catch (InvalidOperationException)
                {
                    continue;
                }
                boardInformation.BoardName = GetBoardName(i, boardType);
                boardInformation.BoardId = GetBoardId(i, boardType);
                boardInformation.BoardInfo = i + ": " + boardInformation.BoardName + " [ID=" + boardInformation.BoardId.ToString() + "]";

                boardInformation.BoardName = GetBoardName(i, boardType);

                // FX3 SSEPM_ENABLEレジスタからSuperSpped or High Speedを取得
                boardInformation.IsUSB3 = ((FpgaRegisterRead(i, boardType, NativeMethods.CHECK_ISUSB3, true) & 0x00008000) >> 15 == 0x1);
                boardInformation.DeviceId = i;
                boardInformation.BoardType = boardType;
                                
                // コレクションへ追加
                boardInformationList.Add(boardInformation);

            }

            return boardInformationList;
        }


        /// <summary>
        /// </summary>
        /// <param name="deviceId"></param>
        /// <param name="boardType"></param>
        /// <returns></returns>
        public static string GetBoardName(int deviceId, int boardType)
        {
            string boardName;

            // Board ID取得処理
            int[] boardId = new int[1]; //IntPtrの相互変換の都合上int[]に
            uint name_length = 256;
            
            NativeMethods.Init();
            
            var buff_name = new StringBuilder(256);

            IntPtr buff_boardid = Marshal.AllocCoTaskMem(Marshal.SizeOf(typeof(int)));
   
            NativeMethods.GetBoardInfo(boardType, deviceId, buff_name, name_length, buff_boardid);
            Marshal.Copy(buff_boardid, boardId, 0,1);
            Marshal.FreeCoTaskMem(buff_boardid);
            
            boardName = buff_name.ToString();

            return boardName;
        }
        
        
        
        /// <summary>
        /// </summary>
        /// <param name="deviceId"></param>
        /// <param name="boardType"></param>
        /// <returns></returns>
        
        public static int GetBoardId(int deviceId, int boardType)
        {
            string boardName;

            // Board ID取得処理
            int[] boardId = new int[1]; //IntPtrの相互変換の都合上int[]に
            uint name_length = 256;
            
            NativeMethods.Init();
            
            var buff_name = new StringBuilder(256);

            IntPtr buff_boardid = Marshal.AllocCoTaskMem(Marshal.SizeOf(typeof(int)));
   
            NativeMethods.GetBoardInfo(boardType, deviceId, buff_name, name_length, buff_boardid);
            Marshal.Copy(buff_boardid, boardId, 0,1);
            Marshal.FreeCoTaskMem(buff_boardid);

            return boardId[0];
        }


        public static byte[] SVMSettingReadArtix(int deviceId, int boardType, UInt32 dataLen, bool boardInitEnd = true)
        {
            UInt32 baseAddr = 0xFF0000;
            byte ucCommdId = 0x03;
            byte ucSSId = 0xE7;
            UInt32 sz = 256;
            byte[] receiveBuffer = new byte[dataLen];
            uint cnt = 0;
            UInt32 ret;

            // boardInitEndがtrueの場合、ボード指定で初期化処理
            if (boardInitEnd == true)
            {
                DeviceOpen(deviceId, boardType);
            }

            while(dataLen > 0)
            {
                if (sz > dataLen) { sz = dataLen; }

                // 受信データ格納用ポインタ
                IntPtr receivePointer = Marshal.AllocCoTaskMem((int)sz);

                ret = NativeMethods.SPIRead(ucCommdId, ucSSId, baseAddr, sz, receivePointer);

                // アンマネージドメモリポインター(DLL側)からマネージド配列(C#側)にデータをコピー
                Marshal.Copy(receivePointer, receiveBuffer, (int)(256 * cnt), (int)sz);

                // 受信データ格納用メモリ空間を開放
                Marshal.FreeCoTaskMem(receivePointer);

                // オフセット
                baseAddr += sz;
                dataLen -= sz;
                cnt++;
            }

            // boardInitEndがtrueの場合、接続済みデバイスのクローズ処理
            if (boardInitEnd == true)
            {
                DeviceClose();
            }

            return receiveBuffer;
        }
        public static void RestartBoard()
        {
            NativeMethods.Restart();   
            DeviceClose();
        }
        
    }
}
